MODULE step
   !!======================================================================
   !!                       ***  MODULE step  ***
   !! Time-stepping    : manager of the ocean, tracer and ice time stepping
   !!======================================================================
   !! History :  OPA  !  1991-03  (G. Madec)  Original code
   !!             -   !  1991-11  (G. Madec)
   !!             -   !  1992-06  (M. Imbard)  add a first output record
   !!             -   !  1996-04  (G. Madec)  introduction of dynspg
   !!             -   !  1996-04  (M.A. Foujols)  introduction of passive tracer
   !!            8.0  !  1997-06  (G. Madec)  new architecture of call
   !!            8.2  !  1997-06  (G. Madec, M. Imbard, G. Roullet)  free surface
   !!             -   !  1999-02  (G. Madec, N. Grima)  hpg implicit
   !!             -   !  2000-07  (J-M Molines, M. Imbard)  Open Bondary Conditions
   !!   NEMO     1.0  !  2002-06  (G. Madec)  free form, suppress macro-tasking
   !!             -   !  2004-08  (C. Talandier) New trends organization
   !!             -   !  2005-01  (C. Ethe) Add the KPP closure scheme
   !!             -   !  2005-11  (G. Madec)  Reorganisation of tra and dyn calls
   !!             -   !  2006-01  (L. Debreu, C. Mazauric)  Agrif implementation
   !!             -   !  2006-07  (S. Masson)  restart using iom
   !!            3.2  !  2009-02  (G. Madec, R. Benshila)  reintroduicing z*-coordinate
   !!             -   !  2009-06  (S. Masson, G. Madec)  TKE restart compatible with key_cpl
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   stp            : OPA system time-stepping
   !!----------------------------------------------------------------------
   USE oce             ! ocean dynamics and tracers variables
   USE dom_oce         ! ocean space and time domain variables 
   USE zdf_oce         ! ocean vertical physics variables
   USE ldftra_oce      ! ocean tracer   - trends
   USE ldfdyn_oce      ! ocean dynamics - trends
   USE in_out_manager  ! I/O manager
   USE iom             !
   USE lbclnk

   USE daymod          ! calendar                         (day     routine)

   USE dtatem          ! ocean temperature data           (dta_tem routine)
   USE dtasal          ! ocean salinity    data           (dta_sal routine)
   USE sbcmod          ! surface boundary condition       (sbc     routine)
   USE sbcrnf          ! surface boundary condition: runoff variables
   USE sbccpl          ! surface boundary condition: coupled formulation (call send at end of step)
   USE cpl_oasis3, ONLY : lk_cpl

#if defined key_top
   USE trcstp          ! passive tracer time-stepping      (trc_stp routine)
#endif

   USE traqsr          ! solar radiation penetration      (tra_qsr routine)
   USE trasbc          ! surface boundary condition       (tra_sbc routine)
   USE trabbc          ! bottom boundary condition        (tra_bbc routine)
   USE trabbl          ! bottom boundary layer            (tra_bbl routine)
   USE tradmp          ! internal damping                 (tra_dmp routine)
   USE traadv          ! advection scheme control     (tra_adv_ctl routine)
   USE traldf          ! lateral mixing                   (tra_ldf routine)
   USE cla             ! cross land advection             (tra_cla routine)
   !   zdfkpp          ! KPP non-local tracer fluxes      (tra_kpp routine)
   USE trazdf          ! vertical mixing                  (tra_zdf routine)
   USE tranxt          ! time-stepping                    (tra_nxt routine)
   USE tranpc          ! non-penetrative convection       (tra_npc routine)

   USE eosbn2          ! equation of state                (eos_bn2 routine)

   USE dynadv          ! advection                        (dyn_adv routine)
   USE dynbfr          ! Bottom friction terms            (dyn_bfr routine)
   USE dynvor          ! vorticity term                   (dyn_vor routine)
   USE dynhpg          ! hydrostatic pressure grad.       (dyn_hpg routine)
   USE dynldf          ! lateral momentum diffusion       (dyn_ldf routine)
   USE dynzdf          ! vertical diffusion               (dyn_zdf routine)
   USE dynspg_oce      ! surface pressure gradient        (dyn_spg routine)
   USE dynspg          ! surface pressure gradient        (dyn_spg routine)
   USE dynnxt          ! time-stepping                    (dyn_nxt routine)

   USE obc_par         ! open boundary condition variables
   USE obcdta          ! open boundary condition data     (obc_dta routine)
   USE obcrst          ! open boundary cond. restart      (obc_rst routine)
   USE obcrad          ! open boundary cond. radiation    (obc_rad routine)

   USE bdy_par         ! unstructured open boundary data variables
   USE bdydta          ! unstructured open boundary data  (bdy_dta routine)

   USE sshwzv          ! vertical velocity and ssh        (ssh_wzv routine)

   USE ldfslp          ! iso-neutral slopes               (ldf_slp routine)
   USE ldfeiv          ! eddy induced velocity coef.      (ldf_eiv routine)

   USE zdftmx          ! tide-induced vertical mixing     (zdf_tmx routine)
   USE zdfbfr          ! bottom friction                  (zdf_bfr routine)
   USE zdftke_old      ! old TKE vertical mixing      (zdf_tke_old routine)
   USE zdftke          ! TKE vertical mixing              (zdf_tke routine)
   USE zdfkpp          ! KPP vertical mixing              (zdf_kpp routine)
   USE zdfddm          ! double diffusion mixing          (zdf_ddm routine)
   USE zdfevd          ! enhanced vertical diffusion      (zdf_evd routine)
   USE zdfric          ! Richardson vertical mixing       (zdf_ric routine)
   USE zdfmxl          ! Mixed-layer depth                (zdf_mxl routine)

   USE zpshde          ! partial step: hor. derivative     (zps_hde routine)

   USE diawri          ! Standard run outputs             (dia_wri routine)
   USE trdicp          ! Ocean momentum/tracers trends    (trd_wri routine)
   USE trdmld          ! mixed-layer trends               (trd_mld routine)
   USE trdmld_rst      ! restart for mixed-layer trends
   USE trdmod_oce      ! ocean momentum/tracers trends
   USE trdvor          ! vorticity budget                 (trd_vor routine)
   USE diagap          ! hor. mean model-data gap         (dia_gap routine)
   USE diahdy          ! dynamic height                   (dia_hdy routine)
   USE diaptr          ! poleward transports              (dia_ptr routine)
   USE diaar5          ! AR5 diagnosics                   (dia_ar5 routine)
   USE diahth          ! thermocline depth                (dia_hth routine)
   USE diafwb          ! freshwater budget                (dia_fwb routine)
   USE diaobs          ! obs-minus-model (assimilation)   (dia_obs routine)
   USE flo_oce         ! floats variables
   USE floats          ! floats computation               (flo_stp routine)

   USE stpctl          ! time stepping control            (stp_ctl routine)
   USE restart         ! ocean restart                    (rst_wri routine)
   USE prtctl          ! Print control                    (prt_ctl routine)

#if defined key_agrif
   USE agrif_opa_sponge ! Momemtum and tracers sponges
#endif

   IMPLICIT NONE
   PRIVATE

   PUBLIC   stp        ! called by opa.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "zdfddm_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/OPA 3.2 , LOCEAN-IPSL (2009) 
   !! $Id$
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt) 
   !!----------------------------------------------------------------------

CONTAINS

#if defined key_agrif
   SUBROUTINE stp( )
      INTEGER             ::   kstp   ! ocean time-step index
#else
   SUBROUTINE stp( kstp )
      INTEGER, INTENT(in) ::   kstp   ! ocean time-step index
#endif
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE stp  ***
      !!                      
      !! ** Purpose : - Time stepping of OPA (momentum and active tracer eqs.)
      !!              - Time stepping of LIM (dynamic and thermodynamic eqs.)
      !!              - Tme stepping  of TRC (passive tracer eqs.)
      !! 
      !! ** Method  : -1- Update forcings and data  
      !!              -2- Update ocean physics 
      !!              -3- Compute the t and s trends 
      !!              -4- Update t and s 
      !!              -5- Compute the momentum trends
      !!              -6- Update the horizontal velocity
      !!              -7- Compute the diagnostics variables (rd,N2, div,cur,w)
      !!              -8- Outputs and diagnostics
      !!----------------------------------------------------------------------
      INTEGER ::   jk       ! dummy loop indice
      INTEGER ::   indic    ! error indicator if < 0
      !! ---------------------------------------------------------------------

#if defined key_agrif
      kstp = nit000 + Agrif_Nb_Step()
!      IF ( Agrif_Root() .and. lwp) Write(*,*) '---'
!      IF (lwp) Write(*,*) 'Grid Number',Agrif_Fixed(),' time step ',kstp
#endif   
      indic = 1                                       ! reset to no error condition

      IF( kstp /= nit000 )   CALL day( kstp )         ! Calendar (day was already called at nit000 in day_init)

      CALL iom_setkt( kstp )                          ! say to iom that we are at time step kstp
      
      CALL rst_opn( kstp )                            ! Open the restart file

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Update data, open boundaries, surface boundary condition (including sea-ice)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      IF( lk_dtatem  )   CALL dta_tem( kstp )         ! update 3D temperature data
      IF( lk_dtasal  )   CALL dta_sal( kstp )         ! update 3D salinity data
                         CALL sbc    ( kstp )         ! Sea Boundary Condition (including sea-ice)
      IF( lk_obc     )   CALL obc_dta( kstp )         ! update dynamic and tracer data at open boundaries
      IF( lk_obc     )   CALL obc_rad( kstp )         ! compute phase velocities at open boundaries
      IF( lk_bdy     )   CALL bdy_dta( kstp )         ! update dynamic and tracer data at unstructured open boundary

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      !  Ocean dynamics : ssh, wn, hdiv, rot                                 !
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                         CALL ssh_wzv( kstp )         ! after ssh & vertical velocity

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Ocean physics update                (ua, va, ta, sa used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                         CALL bn2( tb, sb, rn2b )     ! before Brunt-Vaisala frequency
                         CALL bn2( tn, sn, rn2  )     ! now    Brunt-Vaisala frequency
      !
      !  VERTICAL PHYSICS   
                         CALL zdf_bfr( kstp )         ! bottom friction
      !                                               ! Vertical eddy viscosity and diffusivity coefficients
      IF( lk_zdfric     )   CALL zdf_ric    ( kstp )  ! Richardson number dependent Kz
      IF( lk_zdftke_old )   CALL zdf_tke_old( kstp )  ! TKE closure scheme for Kz (old scheme)
      IF( lk_zdftke     )   CALL zdf_tke    ( kstp )  ! TKE closure scheme for Kz
      IF( lk_zdfkpp     )   CALL zdf_kpp    ( kstp )  ! KPP closure scheme for Kz
      IF( lk_zdfcst     )   THEN                      ! Constant Kz (reset avt, avm[uv] to the background value)
         avt (:,:,:) = rn_avt0 * tmask(:,:,:)
         avmu(:,:,:) = rn_avm0 * umask(:,:,:)
         avmv(:,:,:) = rn_avm0 * vmask(:,:,:)
      ENDIF
      IF( ln_rnf_mouth ) THEN                         ! increase diffusivity at rivers mouths
         DO jk = 2, nkrnf   ;   avt(:,:,jk) = avt(:,:,jk) + 2.e0 * rn_avt_rnf * rnfmsk(:,:) * tmask(:,:,jk)   ;   END DO
      ENDIF
      IF( ln_zdfevd  )   CALL zdf_evd( kstp )         ! enhanced vertical eddy diffusivity

      IF( lk_zdftmx  )   CALL zdf_tmx( kstp )         ! tidal vertical mixing

      IF( lk_zdfddm .AND. .NOT. lk_zdfkpp )   &
         &               CALL zdf_ddm( kstp )         ! double diffusive mixing
                         CALL zdf_mxl( kstp )         ! mixed layer depth

                                                      ! write tke information in the restart file
      IF( lrst_oce .AND. lk_zdftke )   CALL tke_rst( kstp, 'WRITE' )
      !
      !  LATERAL  PHYSICS 
      !
      IF( lk_ldfslp ) THEN                            ! slope of lateral mixing
                         CALL eos( tb, sb, rhd )                ! before in situ density
         IF( ln_zps )    CALL zps_hde( kstp, tb, sb, rhd,  &    ! Partial steps: before horizontal gradient
            &                              gtu, gsu, gru,  &    ! of t, s, rd at the last ocean level
            &                              gtv, gsv, grv )
                         CALL ldf_slp( kstp, rhd, rn2b )        ! before slope of the lateral mixing
      ENDIF
#if defined key_traldf_c2d
      IF( lk_traldf_eiv )   CALL ldf_eiv( kstp )      ! eddy induced velocity coefficient
#  endif

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! diagnostics and outputs             (ua, va, ta, sa used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      IF( lk_floats  )   CALL flo_stp( kstp )         ! drifting Floats
      IF( lk_diahth  )   CALL dia_hth( kstp )         ! Thermocline depth (20 degres isotherm depth)
      IF( lk_diagap  )   CALL dia_gap( kstp )         ! basin averaged diagnostics
      IF( lk_diahdy  )   CALL dia_hdy( kstp )         ! dynamical heigh diagnostics
      IF( lk_diafwb  )   CALL dia_fwb( kstp )         ! Fresh water budget diagnostics
      IF( lk_diaobs  )   CALL dia_obs( kstp )         ! obs-minus-model (assimilation) diagnostics
      IF( ln_diaptr  )   CALL dia_ptr( kstp )         ! Poleward TRansports diagnostics
      IF( lk_diaar5  )   CALL dia_ar5( kstp )         ! ar5 diag
                         CALL dia_wri( kstp )         ! ocean model: outputs

#if defined key_top
      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Passive Tracer Model
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                         CALL trc_stp( kstp )         ! time-stepping
#endif

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Active tracers                              (ua, va used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                             ta(:,:,:) = 0.e0               ! set tracer trends to zero
                             sa(:,:,:) = 0.e0

                             CALL tra_sbc    ( kstp )       ! surface boundary condition
      IF( ln_traqsr      )   CALL tra_qsr    ( kstp )       ! penetrative solar radiation qsr
      IF( lk_trabbc      )   CALL tra_bbc    ( kstp )       ! bottom heat flux
      IF( lk_trabbl_dif  )   CALL tra_bbl_dif( kstp )       ! diffusive bottom boundary layer scheme
      IF( lk_trabbl_adv  )   CALL tra_bbl_adv( kstp )       ! advective (and/or diffusive) bottom boundary layer scheme
      IF( lk_tradmp      )   CALL tra_dmp    ( kstp )       ! internal damping trends
                             CALL tra_adv    ( kstp )       ! horizontal & vertical advection
      IF( n_cla == 1     )   CALL tra_cla    ( kstp )       ! Cross Land Advection (Update Hor. advection)
      IF( lk_zdfkpp )        CALL tra_kpp    ( kstp )       ! KPP non-local tracer fluxes
                             CALL tra_ldf    ( kstp )       ! lateral mixing
#if defined key_agrif
      IF(.NOT. Agrif_Root()) CALL Agrif_Sponge_tra          ! tracers sponge
#endif
                             CALL tra_zdf    ( kstp )       ! vertical mixing and after tracer fields

      IF( ln_dynhpg_imp  ) THEN                             ! semi-implicit hpg (time stepping then eos)
         IF( ln_zdfnpc   )   CALL tra_npc    ( kstp )            ! update after fields by non-penetrative convection
                             CALL tra_nxt    ( kstp )            ! tracer fields at next time step
                             CALL eos( ta, sa, rhd, rhop )       ! Time-filtered in situ density for hpg computation
         IF( ln_zps      )   CALL zps_hde( kstp, ta, sa, rhd,   &   ! Partial steps: time filtered hor. derivative
            &                                   gtu, gsu, gru,  &   ! of t, s, rd at the bottom ocean level
            &                                   gtv, gsv, grv )
         
      ELSE                                                  ! centered hpg  (eos then time stepping)
                             CALL eos( tn, sn, rhd, rhop )       ! now in situ density for hpg computation
         IF( ln_zps      )   CALL zps_hde( kstp, tn, sn, rhd,   &   ! Partial steps: now horizontal derivative
            &                                   gtu, gsu, gru,  &   ! of t, s, rd at the bottom ocean level
            &                                   gtv, gsv, grv )
         IF( ln_zdfnpc   )   CALL tra_npc    ( kstp )       ! update after fields by non-penetrative convection
                             CALL tra_nxt    ( kstp )       ! tracer fields at next time step
      ENDIF 

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Dynamics                                    (ta, sa used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                               ua(:,:,:) = 0.e0             ! set dynamics trends to zero
                               va(:,:,:) = 0.e0

                               CALL dyn_adv( kstp )         ! advection (vector or flux form)
                               CALL dyn_vor( kstp )         ! vorticity term including Coriolis
                               CALL dyn_ldf( kstp )         ! lateral mixing
#if defined key_agrif
      IF(.NOT. Agrif_Root())   CALL Agrif_Sponge_dyn        ! momemtum sponge
#endif
                               CALL dyn_hpg( kstp )         ! horizontal gradient of Hydrostatic pressure
                               CALL dyn_bfr( kstp )           ! bottom friction   
                               CALL dyn_zdf( kstp )         ! vertical diffusion
                               indic=0
                               CALL dyn_spg( kstp, indic )  ! surface pressure gradient
                               CALL dyn_nxt( kstp )         ! lateral velocity at next time step

                               CALL ssh_nxt( kstp )         ! sea surface height at next time step

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Control and restarts
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                               CALL stp_ctl( kstp, indic )
      IF( indic < 0        )   THEN
                               CALL ctl_stop( 'step: indic < 0' )
                               CALL dia_wri_state( 'output.abort', kstp )
      ENDIF
      IF( kstp == nit000   )   CALL iom_close( numror )     ! close input  ocean restart file
      IF( lrst_oce         )   CALL rst_write    ( kstp )   ! write output ocean restart file
      IF( lk_obc           )   CALL obc_rst_write( kstp )   ! write open boundary restart file

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Trends                              (ua, va, ta, sa used as workspace)
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      IF( nstop == 0 ) THEN                         
         IF( lk_trddyn     )   CALL trd_dwr( kstp )         ! trends: dynamics 
         IF( lk_trdtra     )   CALL trd_twr( kstp )         ! trends: active tracers
         IF( lk_trdmld     )   CALL trd_mld( kstp )         ! trends: Mixed-layer 
         IF( lk_trdvor     )   CALL trd_vor( kstp )         ! trends: vorticity budget
      ENDIF

      !>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      ! Coupled mode
      !<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      IF( lk_cpl           )   CALL sbc_cpl_snd( kstp )     ! coupled mode : field exchanges
      !
      !
   END SUBROUTINE stp

   !!======================================================================
END MODULE step
